"""
TODO: Documentation.
"""
from __future__ import annotations
import numpy as np
from typing import Optional, overload
from EasyNN.model.abc import Model, base_array
from EasyNN.model.bias import Bias
from EasyNN.model.weight import Weight
from EasyNN.typing import Array1D


class DenseLayer(Model):
    """
    Create a Dense Layer with an amount of neurons.
    Dense Layers are a combination of a Weight followed by a Bias.
    """
    __parameters: Array1D
    __derivatives: Array1D
    neurons: int
    weight: Weight
    bias: Bias

    def __init__(self: DenseLayer, neurons: int) -> None:
        if not isinstance(neurons, int):
            raise TypeError(f"expected int for neurons, got {type(neurons).__name__} instead")
        elif neurons <= 0:
            raise ValueError(f"neurons must be a positive integer, not {neurons}")
        self.neurons = neurons
        self.weight = Weight(neurons)
        self.bias = Bias(neurons)

    @property
    def last_layer(self: DenseLayer) -> Model:
        """Returns the last layer of the model."""
        return self.layers[-1].last_layer

    @property
    def _parameters(self: DenseLayer) -> Array1D:
        """Provides views into the weight/bias parameters."""
        return self.__parameters

    @_parameters.setter
    def _parameters(self: DenseLayer, parameters: Array1D) -> None:
        self.__parameters = parameters
        self.weight._parameters = parameters[:-self.neurons]
        self.bias._parameters = parameters[-self.neurons:]

    @property
    def _derivatives(self: DenseLayer) -> Array1D:
        """Provides views into the weight/bias derivatives."""
        return self.__derivatives

    @_derivatives.setter
    def _derivatives(self: DenseLayer, derivatives: Array1D) -> None:
        self.__derivatives = derivatives
        self.weight._derivatives = derivatives[:-self.neurons]
        self.bias._derivatives = derivatives[-self.neurons:]

    @property
    def layers(self: DenseLayer) -> tuple[Weight, Bias]:
        """Returns the (weight, bias) layers as a tuple."""
        return (self.weight, self.bias)

    @overload
    def __backward__(self: DenseLayer, dy: Array1D, y: Optional[Array1D] = ..., use_y: bool = ...) -> Array1D:
        ...

    @overload
    def __backward__(self: DenseLayer, dy: Array2D, y: Optional[Array2D] = ..., use_y: bool = ...) -> Array2D:
        ...

    def __backward__(self, dy, y=None, use_y=False):
        return self.weight.backward(self.bias.backward(dy))

    @overload
    def __forward__(self: DenseLayer, x: Array1D) -> Array1D:
        ...

    @overload
    def __forward__(self: DenseLayer, x: Array2D) -> Array2D:
        ...

    def __forward__(self, x):
        result = self.bias(self.weight(x))
        self.__setup__()
        return result

    def __setup__(self: DenseLayer) -> None:
        # Update the parameters every layer is setup,
        # but some are not setup to be the same as the DenseLayer parameters.
        if (
            all(hasattr(layer, "parameters") for layer in self.layers)
            and any(base_array(getattr(self, "parameters", None)) is not base_array(layer.parameters) for layer in self.layers)
        ):
            self._parameters = np.append(self.weight.parameters, self.bias.parameters)
        # Update the derivatives every layer is setup,
        # but some are not setup to be the same as the DenseLayer derivatives.
        if (
            all(hasattr(layer, "derivatives") for layer in self.layers)
            and any(base_array(getattr(self, "derivatives", None)) is not base_array(layer.derivatives) is not None for layer in self.layers)
        ):
            self._derivatives = np.append(self.weight.derivatives, self.bias.derivatives)



"""
TODO: Documentation.
"""
from __future__ import annotations
import numpy as np
from typing import Optional, overload
from EasyNN.model.abc import Model_1D
from EasyNN.typing import Array1D, Array2D


class Weight(Model_1D):
    """
    A Weight layer model, which matrix-multiplies its input by a matrix.

    Use Weight(out_neurons, /) to have the specified number of neurons as output.
    Use Weight(in_neurons, out_neurons, /) to specify the number of inputs as well.

    If only the out_neurons is specified, then the in_neurons is taken from the first input.
    Once the in_neurons is set, it cannot be changed.
    """
    neurons: int

    def __init__(self: Weight, *neurons: int) -> None:
        # Only allow integer inputs.
        if not all(isinstance(n, int) and n > 0 for n in neurons[:2]):
            raise TypeError("neurons must be positive integers")
        # Weight(10), collects the input into 10 neurons.
        elif len(neurons) == 1:
            self.neurons = neurons[0]
        # Weight(128, 10), collects 128 inputs into 10 neurons.
        elif len(neurons) == 2:
            self.x = np.empty(neurons[0])
            self.neurons = neurons[1]
            self.__setup__()
        # Invalid number of inputs.
        else:
            raise TypeError(f"__init__ expect 1-2 positional arguments but got {len(neurons)} instead")

    def __forward__(self: Weight, x: Array1D) -> Array1D:
        return x @ self.matrix.T

    def __backward__(self: Weight, dy: Array1D, y: Optional[Array1D] = None, use_y: bool = False) -> Array1D:
        if isinstance(dy, Array1D):
            self.derivatives = np.outer(dy, self.x)
        else:
            self.derivatives = dy.T @ self.x
            self.derivatives /= dy.shape[-1]
        return dy @ self.matrix

    def __setup__(self: Weight) -> None:
        if not hasattr(self, "parameters"):
            std = np.sqrt(6 / (self.neurons + self.x.shape[-1]))
            self.parameters = np.random.uniform(-std, std, self.neurons * self.x.shape[-1])
        if not hasattr(self, "derivatives"):
            self.derivatives = np.empty(self.neurons * self.x.shape[-1], dtype=float)

    @property
    def matrix(self: Weight) -> Array2D:
        return self.parameters.reshape(self.neurons, -1)

    @matrix.setter
    def matrix(self: Weight, matrix: Array2D) -> None:
        matrix = np.asarray(matrix, dtype=float)
        if not isinstance(matrix, (Array2D[float], Array2D[int])):
            raise ValueError("matrix must be a 2D array of floats or ints")
        self.parameters = matrix.flatten()



"""
TODO: Not complete.
"""
from __future__ import annotations
from EasyNN.loss.abc import Loss
from EasyNN.typing import Array1D


class MeanSquareError(Loss):
    """
    Computes the loss as the mean squared error between the predicted
    output and the true output.
    """

    def __call__(self: MeanSquareError, x: Array1D, y: Array1D) -> float:
        y_pred = self.model(x)
        y_pred -= y
        y_pred **= 2
        return y_pred.mean() / 2

    def dy(self: MeanSquareError, y: Array1D, y_pred: Array1D) -> Array1D:
        y_pred -= y
        y_pred /= y.size
        return y_pred



from dataclasses import dataclass, field
#from EasyNN.model.abc import Model


@dataclass
class ReachValidationAccuracy:
    """
    ReachValidationAccuracy(limit: float = 0.90, patience: int = 1)

    Callback used to terminate when the validation accuracy reaches a specified limit.

    Parameters:
        limit:
            The accuracy limit desired. By default, aims for 90% validation accuracy.
        patience:
            The number of times the accuracy limit must be reached in a row.
            Prevents "lucky" termination.

    Callbacks:
        on_validation_start:
            Checks the validation accuracy on this callback step.

    Example:
        Terminate when 80% validation accuracy is reached.
        >>> from EasyNN.examples.mnist.numbers.untrained import model
        >>> model.callback(ReachValidationAccuracy(limit=0.80))
        >>> model.train()
    """

    limit: float = 0.90
    patience: int = 1
    successes: int = field(default=0, init=False)

    def on_validation_start(self, model) -> None:
        if model.accuracy(*model.validation.sample) > self.limit:
            self.successes += 1
        else:
            self.successes = 0
        model.stop_training |= self.successes >= self.patience



"""
TODO: Not complete.
"""
from __future__ import annotations
import numpy as np
from EasyNN.loss.abc import Loss
from EasyNN.typing import Array1D
import EasyNN.model.abc


class NegativeLikelihood(Loss):
    """
    Computes the loss as the log-likelihood of the desired result.

    Similar to the LogLikelihood, except doesn't take the log(). Leaves
    the log() step to the LogSoftMax, instead of using the SoftMax.

    Note: This should be used with the LogSoftMax activation as the last layer.
    """

    def __call__(self: NegativeLikelihood, x: Array1D, y: Array1D) -> float:
        # Check the inputs.
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)
        if y.ndim > 2:
            raise IndexError(
                "y must be either 0-D (the index of a single sample),\n"
                "    or 1-D (multiple indexes or an array of probabilities),\n"
                "    or 2-D (multiple arrays of probabilities),\n"
                "    but not {y.ndim}-D"
            )
        # Extract the predicted probabilities.
        y_pred = self.model(x)
        # If y is an int, it represents the index of the probability being checked.
        if y.ndim == 0 and y_pred.ndim == 1:
            likelihoods = y_pred[int(y)]
        # If y is an array of indexes, check each index.
        elif y.ndim == 1 and y_pred.ndim == 2:
            likelihoods = y_pred[np.arange(len(y)), y.astype(int)]
        # If y is an array of probabilities, check the largest probability.
        elif y.ndim == 1 and y_pred.ndim == 1:
            likelihoods = y_pred[np.argmax(y.astype(int))]
        # If y is multiple arrays of probabilities, check the largest probability for each row.
        elif y.ndim == 2 and y_pred.ndim == 2:
            likelihoods = y_pred[np.arange(len(y)), np.argmax(y.astype(int), axis=1)]
        # Invalid dimensions.
        else:
            raise IndexError(f"mismatching shapes: y.shape = {y.shape} and y_pred.shape = {y_pred.shape}")
        # return -y.
        return -np.asarray(likelihoods, dtype=float).mean()

    def backward(self: NegativeLikelihood, y: Array1D, y_pred: Array1D, model: EasyNN.model.abc.Model = None) -> Any:
        """
        Updates self.derivatives (e.g. via backpropagation) for optimizers (e.g. gradient descent).
        If the last layer is the LogSoftMax activation, then a special algorithm is used.

        Parameters:
            y: Array1D
                The expected output of the model.
            y_pred: Array1D
                The predicted output from the model.
            model: Model
                The model being backpropagated into.

        Example:
            >>> model.backward_sample(x, y)           # For the user.
            >>> model.backward(loss.dy(y, model(x)))  # How it's implemented.
        """
        from EasyNN.model.activation.log_softmax import LogSoftMax
        use_y = isinstance(model.layers[-1], LogSoftMax)
        if model is None:
            model = self.model
        return model.backward(y_pred if use_y else self.dy(y, y_pred), y, use_y=use_y)

    def dy(self: NegativeLikelihood, y: Array1D, y_pred: Array1D) -> Array1D:
        # Check the inputs.
        y = np.array(y, copy=True)
        if y.ndim > 2:
            raise IndexError(
                "y must be either 0-D (the index of a single sample),\n"
                "    or 1-D (multiple indexes or an array of probabilities),\n"
                "    or 2-D (multiple arrays of probabilities),\n"
                "    but not {y.ndim}-D"
            )
        # Assumes y == 0 if it is not the answer and y == 1 if it is.
        # result = -1 if y matches the sample, else 0.
        result = np.zeros_like(y_pred)
        # y = [arrays of indexes] e.g. y = [2, 1].
        if y.ndim == 1 < y_pred.ndim:
            result[np.arange(len(y)), y] = -1
        # y = one index e.g. y = 2.
        elif y.ndim < y_pred.ndim:
            result[y] = -1
        # y = [arrays of 0 and 1] e.g. y = [0, 0, 1] or y = [[0, 0, 1], [0, 1, 0]].
        else:
            result[y != 0] = -1
        return result



from EasyNN.model import Network, Normalize, Randomize, ReLU, LogSoftMax

# Create the mnist model.
model = Network(
    Normalize(1e-3), Randomize(0.01),
    1024, ReLU,
    256, ReLU,
    10, LogSoftMax,
)


"""
TODO: documentation.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import copy
from heapq import merge
from itertools import count
import numpy as np
import pickle
from typing import Any, Callable, Generator, Generic, Iterator, TypeVar, overload
from EasyNN._abc import AutoDocumentation
from EasyNN.batch.abc import Batch, Dataset
from EasyNN.batch.mini import MiniBatch
from EasyNN.callbacks.print  import Print
from EasyNN.loss.abc import Loss
from EasyNN.loss.mean_square_error import MeanSquareError
from EasyNN.optimizer.abc import Optimizer
from EasyNN.optimizer.adam import Adam
from EasyNN.classifier.classifier import Classifier, LabelType
from EasyNN.typing import Array1D, Array2D, Array3D, ArrayND, Callback, Command, Factory
from EasyNN.utilities.parameters import save, load

Labels = TypeVar("Labels")
ArrayIn = TypeVar("ArrayIn", bound=ArrayND)
ArrayOut = TypeVar("ArrayOut", bound=ArrayND)

@overload
def base_array(x: None) -> None:
    ...

@overload
def base_array(x: ArrayND) -> ArrayND:
    ...

def base_array(x):
    """Returns the base of the numpy array."""
    while x is not None is not x.base:
        x = x.base
    return x


class Model(AutoDocumentation, ABC, Generic[ArrayIn, ArrayOut]):
    """
    TODO: documentation.
    """
    _callbacks: dict[Command, list[Callback]]
    _command: Command = "off"
    _derivatives: Array1D
    _parameters: Array1D
    _x: ArrayIn
    _y: ArrayOut
    _training: Dataset[ArrayIn, ArrayOut]
    _testing: Dataset[ArrayIn, ArrayOut]
    _validation: Dataset[ArrayIn, ArrayOut]
    _classifier: Classifier[Labels]
    _optimizer: Optimizer
    _labels: Labels
    _default_batch: Factory[Batch] = MiniBatch
    _default_loss: Factory[Loss[ArrayIn, ArrayOut]] = MeanSquareError
    _default_optimizer: Factory[Optimizer] = Adam
    _default_classifier: Factory[Classifier] = Classifier
    _print: Print
    stop_training: bool = False

    def __getstate__(self) -> dict[str, Any]:
        """Pickle the non-array variables."""
        attributes = {
            name: attribute
            for name, attribute in vars(self).items()
            if name != "command"
            if name != "commands"
            if not isinstance(attribute, np.ndarray)
        }
        if hasattr(self, "x"):
            attributes["x_shape"] = self.x.shape
        elif hasattr(self, "parameters"):
            attributes["parameters_shape"] = self.parameters.shape
        return attributes

    def save(self, file: str) -> None:
        save(file + "_parameters.npz", **self.get_arrays())
        with open(file + "_structure.pkl", "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(file: str) -> Model:
        with open(file + "_structure.pkl", "rb") as f:
            self = pickle.load(f)
        arrays = load(file + "_parameters.npz")
        self.set_arrays(**{name: array for name, array in arrays.items() if name != "parameters"})
        if hasattr(self, "x_shape"):
            x_shape = vars(self).pop("x_shape")
            self(np.zeros(x_shape))
        elif hasattr(self, "parameters_shape"):
            parameters_shape = vars(self).pop("parameters_shape")
            self.parameters = np.empty(parameters_shape)
        self.set_arrays(**arrays)
        return self

    def _on_optimization_start(self) -> None:
        self(self.training[0][0])

    def _on_training_start(self) -> None:
        next(self.training)

    def _on_testing_start(self) -> None:
        next(self.testing)

    def _on_validation_start(self) -> None:
        next(self.validation)

    @property
    def callbacks(self) -> dict[Command, list[Callback]]:
        """Stores the callback commands."""
        if not hasattr(self, "_callbacks"):
            self._callbacks = defaultdict(list)
            # At the start of optimization, compile the model.
            self.on_optimization_start(type(self)._on_optimization_start)
            # At the start of each callback, get the next batch sample from the datasets.
            self.on_training_start(type(self)._on_training_start)
            self.on_testing_start(type(self)._on_testing_start)
            self.on_validation_start(type(self)._on_validation_start)
        return self._callbacks

    @property
    def command(self) -> Command:
        """Stores the current command being run."""
        return self._command

    @command.setter
    def command(self, command: Command) -> None:
        self._command = command
        if len(self.layers) == 1 and self.layers[0] is self:
            return
        for layer in self.layers:
            layer.command = command

    @property
    def layers(self) -> tuple[Model, ...]:
        """Returns the layers of the model."""
        # By default only one layer. Networks and DenseLayers should override this property.
        return (self,)

    @property
    def derivatives(self) -> Array1D:
        """Model derivatives are a 1D array used to store parameter derivatives during model.backward(dy)."""
        return self._derivatives

    @derivatives.setter
    def derivatives(self, derivatives: Array1D) -> None:
        if hasattr(self, "derivatives"):
            self._derivatives[...] = np.reshape(derivatives, -1)
        else:
            self._derivatives = np.asarray(derivatives, dtype=float).reshape(-1)

    @property
    def parameters(self) -> Array1D:
        """Model parameter values are a 1D array which can be modified to change the model."""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: Array1D) -> None:
        if hasattr(self, "parameters"):
            self._parameters[...] = np.reshape(parameters, -1)
        else:
            self._parameters = np.asarray(parameters, dtype=float).reshape(-1)

    @property
    def batch(self) -> Batch:
        """Used to generate samples from the dataset."""
        if not hasattr(self, "_batch"):
            self.batch = self._default_batch()
        return self._batch

    @batch.setter
    def batch(self, batch: Batch) -> None:
        self._batch = batch

    @property
    def classifier(self) -> Classifier[T]:
        """Used to classify the outputs of the model."""
        if not hasattr(self, "_classifier"):
            self.classifier = self.loss._default_classifier()
        return self._classifier

    @classifier.setter
    def classifier(self, classifier: Classifier[T]) -> None:
        self._classifier = classifier

    @property
    def loss(self) -> Loss[ArrayIn, ArrayOut]:
        """Reference to the loss stored in the optimizer."""
        if not hasattr(self, "_loss"):
            self.loss = self._default_loss()
        return self._loss

    @loss.setter
    def loss(self, loss: Loss[ArrayIn, ArrayOut]) -> None:
        self._loss = copy(loss)
        self.loss.model = self

    @property
    def labels(self) -> dict[int, Labels]:
        return self.training.labels

    @labels.setter
    def labels(self, labels) -> None:
        """Sets the labels when asked for them."""
        self.training.labels = labels

    @property
    def optimizer(self) -> Optimizer:
        """Stores the optimizer being used for training."""
        if not hasattr(self, "_optimizer"):
            self.optimizer = self._default_optimizer()
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer) -> None:
        optimizer.add(self)
        self._optimizer = optimizer

    @property
    def iteration(self) -> int:
        """Returns the current iteration from training."""
        return self.training.iteration

    @property
    def epoch(self) -> int:
        """Returns the current epoch from training."""
        return self.training.epoch

    @property
    def training(self) -> Dataset:
        """
        The training dataset the main dataset used during training.
        Use model.training.data = (x, y) to set the training data.
        """
        if not hasattr(self, "_training"):
            self._training = Dataset()
        return self._training

    @property
    def testing(self) -> Dataset:
        """
        The testing dataset is used to produce unbiased evaluation of the model.
        Use model.tsting.data = (x, y) to set the testing data.
        """
        if not hasattr(self, "_testing"):
            self._testing = Dataset()
        return self._testing

    @property
    def validation(self) -> Dataset:
        """
        The validation dataset is used as a pseudo testing dataset during training.
        Use model.validation.data = (x, y) to set the validation data.
        """
        if not hasattr(self, "_validation"):
            self._validation = Dataset()
        return self._validation

    @property
    def x(self) -> ArrayIn:
        """
        Stores the x values from the previous call for future use,
        such as for backpropagation.
        """
        return self._x

    @x.setter
    def x(self, x: ArrayIn) -> None:
        self._x = np.asarray(x, dtype=float)

    @property
    def y(self) -> ArrayOut:
        """
        Stores the y values from the previous call for future use,
        such as for backpropagation.
        """
        return self._y

    @y.setter
    def y(self, y: ArrayOut) -> None:
        self._y = np.asarray(y, dtype=float)

    @property
    def print(self) -> Print:
        if not hasattr(self, "_print"):
            self._print = Print()
            self.callback(self._print)
        return self._print

    def get_arrays(self) -> dict[str, ArrayND]:
        """Returns the arrays stored in the model."""
        if hasattr(self, "parameters"):
            return dict(parameters=self.parameters)
        else:
            return dict()

    def set_arrays(self, *, parameters: ArrayND = None) -> None:
        """Sets the arrays stored in the model."""
        if parameters is not None:
            self.parameters = parameters

    def prepare_datasets(self) -> None:
        """Prepares the datasets before optimizing."""
        if not hasattr(self.training, "data"):
            raise ValueError("requires model.training.data to be set to train the model")
        elif len(self.training) < 10:
            raise ValueError("requires len(model.training.data) >= 10 to train the model")
        # Shuffle the training dataset.
        self.training.data = self.training[np.random.permutation(len(self.training))]
        # Steal a percent of the testing data from the training data if necessary.
        if not hasattr(self.testing, "data"):
            self.testing.data = self.training[:int(len(self.training) * self.testing.percent)]
            self.training.data = self.training[int(len(self.training) * self.testing.percent):]
        # Steal a percent of the validation data from the training data if necessary.
        if not hasattr(self.validation, "data"):
            self.validation.data = self.training[:int(len(self.training) * self.validation.percent)]
            self.training.data = self.training[int(len(self.training) * self.validation.percent):]
        # Apply the batches to each dataset.
        if not hasattr(self.training, "batch"):
            self.training.batch = self.batch
        if not hasattr(self.testing, "batch"):
            self.testing.batch = MiniBatch(len(self.testing))
        if not hasattr(self.validation, "batch"):
            self.validation.batch = MiniBatch(1024)
        # Apply the datasets to each layer.
        for layer in self.layers:
            layer._training = self.training
            layer._testing = self.testing
            layer._validation = self.validation

    def fit(self, x: ArrayIn, y: ArrayOut) -> None:
        """
        Fit the model to the provided training data.

        Equivalent to setting the training.data, and then using model.train().
        """
        self.training.data = (x, y)
        self.train()

    def train(self) -> None:
        """Train the model."""
        self.optimizer.train(self)

    def callback(self, *callbacks: Callback) -> None:
        """model.callback(...) consumes a callback object.."""
        for cb in callbacks:
            for method in dir(type(cb)):
                if method.startswith("on_"):
                    self.callbacks[method].append(getattr(cb, method))

    def on_optimization_start(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_optimization_start')."""
        for cb in callbacks:
            self.callbacks["on_optimization_start"].append(cb)
        return cb

    def on_optimization_end(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_optimization_end')."""
        for cb in callbacks:
            self.callbacks["on_optimization_end"].append(cb)
        return cb

    def on_training_start(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_training_start')."""
        for cb in callbacks:
            self.callbacks["on_training_start"].append(cb)
        return cb

    def on_training_end(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_training_end')."""
        for cb in callbacks:
            self.callbacks["on_training_end"].append(cb)
        return cb

    def on_testing_start(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_testing_start')."""
        for cb in callbacks:
            self.callbacks["on_testing_start"].append(cb)
        return cb

    def on_testing_end(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_testing_end')."""
        for cb in callbacks:
            self.callbacks["on_testing_end"].append(cb)
        return cb

    def on_validation_start(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_validation_start')."""
        for cb in callbacks:
            self.callbacks["on_validation_start"].append(cb)
        return cb

    def on_validation_end(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_validation_end')."""
        for cb in callbacks:
            self.callbacks["on_validation_end"].append(cb)
        return cb

    def on_epoch_end(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_epoch_end')."""
        for cb in callbacks:
            self.callbacks["on_epoch_end"].append(cb)
        return cb

    def on_epoch_start(self, *callbacks: Callback) -> Callback:
        """Shortcut for model.callback('on_epoch_start')."""
        for cb in callbacks:
            self.callbacks["on_epoch_start"].append(cb)
        return cb

    def training_validation_commands(self) -> Iterator[Command]:
        """Generates the training/validation commands in an even manner."""
        # Set commands = (step, "start"), (step, "end"), (2 * step, "start"), (2 * step, "end"), ...
        # The step is used to control how frequently each command occurs.
        training_step = len(self.validation) * self.training._batch_size
        training_commands = (
            (i, command)
            for i in count(start=training_step, step=training_step)
            for command in ("on_training_start", "on_training_end")
        )
        validation_step = len(self.training) * self.validation._batch_size
        validation_commands = (
            (i, command)
            for i in count(start=validation_step, step=validation_step)
            for command in ("on_validation_start", "on_validation_end")
        )
        # Sort the commands by the steps.
        for step, command in merge(training_commands, validation_commands, key=lambda step_command: step_command[0]):
            yield command

    def _optimizer_commands(self) -> Iterator[Command]:
        """
        Generates all of the commands for the optimizer.

        Follows the following sequence of events:
            optimization_start -> training/validation -> optimization_end -> testing_start -> testing_end
        """
        # Start optimizing.
        yield "on_optimization_start"
        # Track the epoch to know when it changes.
        epoch = self.training.epoch
        yield "on_epoch_start"
        # Loop through the training/validation commands until training stops.
        for command in self.training_validation_commands():
            yield command
            if self.stop_training:
                break
            # Switch between epochs whenever the training epoch changes.
            if -1 < epoch < self.training.epoch:
                yield "on_epoch_end"
                yield "on_epoch_start"
            epoch = self.training.epoch
        # Trigger on_{training/validation}_end and on_epoch_end when training stops.
        if command.endswith("_start"):
            yield command.replace("_start", "_end")
        yield "on_epoch_end"
        # Stop optimizing.
        yield "on_optimization_end"
        # Test the results.
        yield "on_testing_start"
        yield "on_testing_end"
        # Disable training mode.
        yield "off"

    def optimizer_commands(self) -> Iterator[Command]:
        """Generates all of the commands and runs their callbacks for the optimizer."""
        # Prepare the datasets before optimizing.
        self.prepare_datasets()
        # Run all of the commands.
        for self.command in self._optimizer_commands():
            # Run all of the callbacks.
            for callback in self.callbacks[self.command]:
                callback(self)
            yield self.command
        self.command = "off"

    def accuracy(self, x: ArrayIn=None, y: ArrayOut=None) -> float:
        """Returns the classification accuracy of the data."""
        # If the user doesnt specify the dataset. We will try to give it them using the training data
        try:
            return self.classifier.accuracy(self(x), y)
        except:
            return self.classifier.accuracy(self(self.training.data[0]), self.training.data[1])
        else:
            print("Could not load training data to calculate accuracy data.")
            raise
            
    def classify(self, x: ArrayIn) -> T:
        """Returns the classification of the input data."""
        return self.classifier.classify(self(x), self.labels)

    def sample_derivatives(self, x: ArrayIn, y: ArrayOut) -> None:
        """Shortcut for `model.loss.backward(y, model(x), model)`."""
        self.loss.backward(y, self(x), self)

    def __call__(self, x: ArrayIn) -> ArrayOut:
        """
        Implements y = model(x) for feed-forward propagation.
        Parses the input as a numpy array before using the forward implementation method.
        """
        self.x = x
        if not hasattr(self, "parameters"):
            self.__setup__()
        self.y = self.__forward__(self.x)
        if not hasattr(self, "parameters"):
            self.__setup__()
        return self.y

    def backward(self, dy: ArrayOut, y: ArrayOut = None, use_y: bool = False) -> ArrayIn:
        """
        Implements dx = model.backward(dy) for backpropagation.
        Parses the input as a numpy array before using the backward implementation method.
        """
        return self.__backward__(
            dy=np.asarray(dy),
            y=(None if y is None else np.asarray(y)),
            use_y=use_y,
        )

    @abstractmethod
    def __backward__(self, dy: ArrayOut, y: ArrayOut = None, use_y: bool = False) -> ArrayIn:
        """Implements the backpropagation after the input has been parsed."""
        raise NotImplementedError

    @abstractmethod
    def __forward__(self, x: ArrayIn) -> ArrayOut:
        """Implements the feed-forward propagation after the input has been parsed."""
        raise NotImplementedError

    @abstractmethod
    def __setup__(self) -> None:
        """
        Implements the setup procedure for the values, derivatives, shapes in, and shapes
        out.
        """
        raise NotImplementedError


class Model_1D(Model[Array1D, Array1D], ABC):
    """Base model for handling 1D inputs and outputs."""


class Model_2D(Model[Array2D, Array2D], ABC):
    """Base model for handling 2D inputs and outputs."""


class Model_3D(Model[Array3D, Array3D], ABC):
    """Base model for handling 3D inputs and outputs."""



from EasyNN.optimizer.abc import Optimizer
from EasyNN.optimizer.adam import Adam
from EasyNN.optimizer.gradient_descent import GradientDescent
from EasyNN.optimizer.momentum_descent import MomentumDescent
from EasyNN.optimizer.nesterov import Nesterov



"""
Contains all of the type-hint related things.
"""
from nptyping import NDArray, Shape
from numpy.typing import ArrayLike
from typing import Any, Callable, Literal, Sequence, Type, TypeVar, Union

T = TypeVar("T")

Factory = Union[T, Callable[[], T]]
Array1D = NDArray[Shape["*"], Any]
Array2D = NDArray[Shape["*, *"], Any]
Array3D = NDArray[Shape["*, *, *"], Any]
ArrayND = NDArray[Any, Any]
ArrayLike1D = Union[Array1D, Sequence[Any]]
ArrayLike2D = Union[Array2D, Sequence[Sequence[Any]]]
ArrayLike3D = Union[Array3D, Sequence[Sequence[Sequence[Any]]]]
ArrayLikeND = ArrayLike
Sample = tuple[ArrayND, ...]

Callback = Callable[[], Any]
Command = Literal[
    "off",
    "on_optimization_start",
    "on_optimization_end",
    "on_training_start",
    "on_training_end",
    "on_testing_start",
    "on_testing_end",
    "on_validation_start",
    "on_validation_end",
    "on_epoch_start",
    "on_epoch_end",
]



"""
Contains all of the abstract base classes.
"""
from EasyNN.model.abc import Model
from EasyNN.activation.abc import Activation
from EasyNN.optimizer.abc import Optimizer
from EasyNN.batch.abc import Batch
from EasyNN.dataset.abc import Dataset
from EasyNN.loss.abc import Loss



from EasyNN.loss.abc import Loss
from EasyNN.loss.log_likelihood import LogLikelihood
from EasyNN.loss.mean_square_error import MeanSquareError
from EasyNN.loss.negative_likelihood import NegativeLikelihood



from dataclasses import dataclass
from matplotlib import pyplot as plt
import numpy as np
from typing import Literal

from EasyNN.callbacks.plot.abc import Plotter
#from EasyNN.model.abc import Model


@dataclass
class PlotValidationLoss(Plotter):
    """
    PlotValidationLoss(
        lr: float = 0.3,
        name: str = "validation loss",
        x_axis: Literal["epochs", "iterations", "validation", "training"] = "epochs"
    )

    A Plotter callback for plotting the validation loss.

    Parameters:
        lr:
            The rate at which the smoothened graph follows the raw data.
        name:
            The name of the data being plotted (for titles/labels).
        x_axis:
            Set the x-axis to epochs or iterations.
            By default, the epochs are used for the x-axis.
            By default, iterations is the validation iterations.
            May also be set to "training" for the training iterations.

    Example:
        Plot the validation loss for the untrained mnist numbers model.
        >>> from EasyNN.examples.mnist.numbers.untrained import model
        >>> model.callback(PlotValidationLoss())
        >>> model.train()
    """
    name: str = "validation loss"
    x_axis: Literal["epochs", "iterations", "validation", "training"] = "epochs"

    def x_values(self, model) -> np.ndarray:
        if self.x_axis in ("iterations", "validation"):
            return np.arange(len(self.data))
        elif self.x_axis == "training":
            return np.arange(len(self.data)) * model.validation._batch_size / model.training._batch_size
        else:
            return np.arange(len(self.data)) * model.validation._batch_size / len(model.validation)

    def on_validation_start(self, model) -> None:
        self.data.append(model.loss(*model.validation.sample))

    def on_testing_start(self, model) -> None:
        super().on_testing_start(model)
        plt.xlabel(self.x_axis)
        plt.ylabel("loss")


from __future__ import annotations
import numpy as np
from typing import Optional, overload
from EasyNN.model.activation.abc import Activation
from EasyNN.typing import Array1D, Array2D, Array3D, ArrayND


class ReLU(Activation):

    @overload
    def __backward__(self: ReLU, dy: Array1D, y: Optional[Array1D] = ..., use_y: bool = ...) -> Array1D:
        ...

    @overload
    def __backward__(self: ReLU, dy: Array2D, y: Optional[Array2D] = ..., use_y: bool = ...) -> Array2D:
        ...

    def __backward__(self, dy, y=None, use_y=False):
        dy *= np.sign(self.x)
        return dy

    @overload
    def __forward__(self: ReLU, x: Array1D) -> Array1D:
        ...

    @overload
    def __forward__(self: ReLU, x: Array2D) -> Array2D:
        ...

    def __forward__(self, x):
        return np.maximum(x, 0, out=x)



from EasyNN.utilities import Preprocess
import numpy as np
import os

def create_image_dataset(name:str,path:str,size:list[int]=(28,28),greyscale:bool=True):
    """ 
    Create a dataset from images in a folders with folder names as labels.

    Example:
        Images:
            ├── cat <- Folder and label to the images.
            │   ├── 0.jpg <- Image of cat
            │   ├── 1.jpg
            └── dog <- Folder and label to the images.
                ├── 0.jpg <- Image of dog
                ├── 1.jpg <- Image of dog

    >>> create_image_dataset("Animal", "./images/", (28,28))
    >>> dataset = load("Animal_dataset.npz")
    >>> images, labels = dataset
    >>> print(images.shape)
    >>> print(labels.shape)
    (4, 784)
    (4,)

    Shows 4 labels and 4 images to be used in a neural network.

    Arguements:
        name: Name of the of dataset file the data will be saved to.
        path: Path to the folder where your labeled folders are.
        size: Size the images will be resized to.
            Default = (28,28) pixels
        greyscale: Change the image to greyscale values.
            Default = True

    Returns:
        Numpy file with images and labels stored in numpy arrays.
    """
    x = np.zeros(shape = (1, 784), dtype = int)
    y = np.array([], dtype = int)
    labels = np.array([], dtype= str)

    # Each folder in the directory.
    for index,folder in enumerate(os.listdir(path)):

        # List all the files in the folder selected.
        files = os.listdir(path+folder)
        files:list[str]

        # The label is the folder name.
        folder:str

        # Store the folder name into the labels array.
        labels = np.append(labels,folder)

        for file in files:
            format_options = dict(
                grayscale=True,
                resize=size
            )

            # Converting your image into the correct format for the mnist number dataset.
            image = Preprocess(path+folder+'/'+file).format(**format_options)

            x = np.append(x,[image],axis = 0)
            y = np.append(y,index)

    # Removes the first empty array.
    x = np.delete(x,0,0)
    # Save the file.
    np.savez(name+"_dataset", x=x, y=y)


"""
TODO: Not complete.
"""
from __future__ import annotations
from typing import Any
from EasyNN.optimizer.gradient_descent import GradientDescent
from EasyNN.utilities.momentum import Momentum
import EasyNN.model.abc


class MomentumDescent(GradientDescent):
    """Momentum Descent uses parameters -= lr * momentum."""
    _momentum_lr: float

    def __init__(self: MomentumDescent, *args: Any, momentum_lr: float = 1e-2, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._momentum_lr = momentum_lr

    def setup(self: MomentumDescent, model: EasyNN.model.abc.Model) -> None:
        """Setup a model."""
        super().setup(model)
        model._derivative_momentum = Momentum(self._momentum_lr)

    def get_derivatives(self: MomentumDescent, model: EasyNN.model.abc.Model) -> Array1D[float]:
        """Computes the derivatives for the optimizer."""
        if model.command.startswith("on_training"):
            return model._derivative_momentum.update(super().get_derivatives(model))
        else:
            return model._derivative_momentum.peek(super().get_derivatives(model))



from EasyNN.examples.cifar10 import labels, show
from EasyNN.model import Network, Normalize, Randomize, ReLU, LogSoftMax
from EasyNN.optimizer import MomentumDescent
from EasyNN.batch import MiniBatch
import EasyNN.callbacks as cb


# Create the mnist model.
model = Network(
    Normalize(1e-6),
    Randomize(0.3), 1024, ReLU,
    Randomize(0.2), 256, ReLU,
    Randomize(0.1), 256, ReLU,
    Randomize(0.03), 10,  LogSoftMax
)

# Set the validation batch size.
model.validation.batch = MiniBatch(256)

# Establish the labels and show feature.
model.labels = labels
model.show = show

# Use gradient descent with momentum.
model.optimizer = MomentumDescent()

# Change the default learning rate.
model.optimizer.lr = 0.03

model.callback(
    # Set when to terminate point. 
        # In this case it will end once your validation accuracy hits above 90% five times.
    cb.ReachValidationAccuracy(limit=0.90, patience=2),
)

# When the model hit a validation point it will print the iteration and accuracy of the model.
model.print.on_validation_start(iteration=True, accuracy=True)
# When the model completes 10 iterations. It will print that iteration number.
model.print.on_training_start(iteration=True, frequency=10)



from EasyNN.callbacks.plot.abc import Plotter
from EasyNN.callbacks.plot.validation_accuracy import PlotValidationAccuracy
from EasyNN.callbacks.plot.validation_loss import PlotValidationLoss
from EasyNN.callbacks.plot.training_accuracy import PlotTrainingAccuracy


from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import matplotlib.pyplot as plt
import numpy as np

#from EasyNN.model.abc import Model
from EasyNN.utilities.momentum import Momentum


@dataclass
class Plotter:
    lr: float = 0.3
    name: str = "data"
    data: list = field(default_factory=list, init=False)

    @abstractmethod
    def x_values(self, model) -> np.ndarray:
        """Rescales the iteration into another format (e.g. validation iteration, training iteration, or epoch)."""

    def on_testing_start(self, model) -> None:
        x = self.x_values(model)
        y = self.data
        plt.plot(x, y, label=f"raw {self.name}")
        smoothener = Momentum(self.lr)
        y_smooth = [smoothener.update(data) for data in y]
        plt.plot(x, y_smooth, label=f"smooth {self.name}")
        plt.title(self.name.title())
        plt.legend(loc="lower left")

    def on_testing_end(self, model) -> None:
        plt.show()



from EasyNN.model.activation.abc import Activation
from EasyNN.model.activation.log_softmax import LogSoftMax
from EasyNN.model.activation.relu import ReLU
from EasyNN.model.activation.softmax import SoftMax



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Iterator
from EasyNN._abc import AutoDocumentation
from EasyNN.dataset import Dataset
from EasyNN.typing import Sample


class Batch(AutoDocumentation, ABC):
    """
    Abstract Base Class for generating samples to loop over from the dataset.
    """
    size: int

    @abstractmethod
    def generate_samples(self: Batch, dataset: Dataset) -> Iterator[Sample]:
        """Generates samples to loop over."""
        raise NotImplementedError



import numpy as np
from typing import TypeVar
from EasyNN.typing import ArrayND

ArrayOut = TypeVar("ArrayOut")
LabelType = TypeVar("LabelType")


class Classifier:
    """Used for giving users the ability to tell model accuracy using
     validation data from the dataset. It also and classify images using
      the predictions and labels."""

    labels: list[LabelType]
    
    def accuracy(self, y_pred: ArrayOut, y: ArrayOut)-> float:
        """
        Used to tell the user how accurate the model is using a portion
        of the dataset that the model has never seen.

        Args:
            validation_data: A batch of validation_data (data,label) used to check the model.
        Return:
            Return a float that is used to establish how good the model is.
        Example:
            >>> model.accuracy(dataset)
            0.92
        """

        correct = 0

        for predictions,label in zip(y_pred,y):
            if np.argmax(predictions) == label:
                correct += 1
        
        return correct / len(y_pred)

        

    def classify(self, y_pred: ArrayOut, labels) -> LabelType:
        """
        Takes the labels and prediction from the model and 
        
        """
        return labels[np.argmax(y_pred)]

        


from EasyNN.utilities.download import download
from EasyNN.utilities.data.load import load
from . import file, url
from os.path import exists

if not exists(file):
    download(file, url)
dataset = load(file)





import numpy as np
from os.path import exists
from EasyNN.utilities.download import download

def load(file_path: str, url: str = None):
    """Used for loading model files that have been downloaded.
    
    Args:
        file_path: Path to file to be loaded.
        url (optional): If you also want to download the file
        
    Returns:
        Returns a numpy list of all variables stored in the file.

    Example:
        >>> data = load("model/mnist.npz")
        >>> print(list(data.keys()))
        ['parameters', 'mean', 'variance']
    """
    # If it has already been downloaded skip.
    if not exists("file_path"):
        # If the file doesnt exist but user has given a url.
        if url:
            download(file_path, url)
    # Load the data.
    return np.load(file_path)


def save(file_name: str, *args, **kwargs) -> None:
    """Used for saving model parameters that can been downloaded/used later.
    
    Args:
        file_name: File name is to be saved under.
        variables: numpy arrays to be saved.

    Example:

        Save variables to file named "model.npz".
        >>> variables = dict(
        >>>    parameters = model.parameters,
        >>>    mean = model.mean,
        >>>    variance = model.variance
        >>> )
        >>> save("model", **variables)
    """
    np.savez_compressed(file_name, *args, **kwargs)
    print("Parameters saved.")  
 


"""
TODO: Not complete.
"""
from __future__ import annotations
from EasyNN.optimizer.momentum_descent import MomentumDescent
import EasyNN.model.abc


class Nesterov(MomentumDescent):
    """Nesterov uses parameters -= lr * momentum, where the momentum is computed by looking ahead."""

    def get_derivatives(self: Nesterov, model: EasyNN.model.abc.Model) -> Array1D[float]:
        """Computes the derivatives for the optimizer."""
        if model.training.iteration == 0:
            return super().get_derivatives(model)
        parameters = model.parameters.copy()
        self.on_training_start(model, model._derivative_momentum.value)
        super().get_derivatives(model)
        model.parameters = parameters
        return model._derivative_momentum.value



"""
TODO: Not complete.
"""
from __future__ import annotations
import numpy as np
from typing import Any
from EasyNN.optimizer.abc import Optimizer
from EasyNN.optimizer.momentum_descent import MomentumDescent
from EasyNN.utilities.momentum import Momentum
import EasyNN.model.abc


class Adam(MomentumDescent):
    """Adam optimizer uses parameters -= lr * momentum / L2_norm."""
    _L2_lr: float
    _L2_epsilon: float

    def __init__(self: MomentumDescent, *args: Any, L2_lr: float = 1e-3, L2_epsilon: float = 1e-7, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._L2_lr = L2_lr
        self._L2_epsilon = L2_epsilon

    def setup(self: MomentumDescent, model: EasyNN.model.abc.Model) -> None:
        """Setup a model."""
        super().setup(model)
        model._L2_momentum = Momentum(self._L2_lr)
        model._L2_epsilon = self._L2_epsilon

    def get_derivatives(self: MomentumDescent, model: EasyNN.model.abc.Model) -> Array1D[float]:
        """Computes the derivatives for the optimizer."""
        derivatives = Optimizer.get_derivatives(self, model)
        return model._derivative_momentum.update(derivatives) / np.sqrt(model._L2_momentum.update(derivatives ** 2) + model._L2_epsilon)



from EasyNN.examples.mnist.number.structure import model
from EasyNN.examples.mnist.number.data import dataset
from EasyNN.examples.mnist.number import labels, show
from EasyNN.optimizer import MomentumDescent
import EasyNN.callbacks as cb
import numpy as np

# Set your models data
model.training.data = dataset

# Establish the labels and show feature.
model.labels = labels
model.show = show 

# Use gradient descent with momentum.
model.optimizer = MomentumDescent()

# Change the default learning rate.
model.optimizer.lr = 0.03

model.callback(
    # Set when to terminate point. 
        # In this case it will end once your validation accuracy hits above 90% five times.
    cb.ReachValidationAccuracy(limit=0.95, patience=2),
)

# When the model hit a validation point it will print the iteration and accuracy of the model.
model.print.on_validation_start(iteration=True,accuracy=True)
# When the model completes 10 iterations. It will print that iteration number.
model.print.on_training_start(iteration=True, frequency=10)




from __future__ import annotations
import numpy as np
from EasyNN.model.activation.abc import Activation
from EasyNN.loss.log_likelihood import LogLikelihood
from EasyNN.typing import Array1D


class SoftMax(Activation):
    _default_loss = LogLikelihood

    def __backward__(self: SoftMax, dy: Array1D, y: Array1D = None, use_y: bool = False) -> Array1D:
        if not use_y:
            dy *= self.y
            dy -= dy.sum(axis=-1, keepdims=True) * self.y
            return dy
        elif y.shape == dy.shape:
            self.y -= y
            return self.y
        elif y.ndim == 0:
            self.y[y] -= 1
            return self.y
        else:
            self.y[np.arange(len(y)), y] -= 1
            return self.y

    def __forward__(self: SoftMax, x: Array1D) -> Array1D:
        self.y = x.copy()
        self.y -= self.y.max(axis=-1, keepdims=True)
        self.y = np.exp(self.y, out=self.y)
        self.y /= self.y.sum(axis=-1, keepdims=True)
        return self.y



import matplotlib.pyplot as plt
import random

__all__ = ["compare"]

def compare(*args, width=3, height=3, dataset=None) -> None:
    """
    Used to compare matplotlib images to eachother.

    Args:
        *args: All the images given to the function
        width: Used to tell the maximum images you want on one row.
        dataset(x,y): If the user wants to compare their image
        to a bunch of dataset images they can feed in the whole
        dataset and the remaing image spot will be filled. 

    Return:
        Subplot of all the images.

    Example:
        >>> compare(image1,image2,image3)
        Returns subplot of image1,image2,image3

        >>> compare(image, dataset = dataset)
        Returns subplot of users image and the rest fill in
        from the datasets images.
    """

    # Check if all the images are the same size.
        # If not raise - Images are not in the same format or size.

    # Maybe this should also be a model function to add simplicity to the function.
        # This is because each one changes its requirements based on the type of the image.

    args_count = len(args)
        
    for i in range(args_count):
        # define subplot
        plt.subplot(width * 110 + 1 + i)
        # plot raw pixel data
        plt.imshow(args[i].reshape([28, 28]), cmap=plt.get_cmap('gray'))
    
    if dataset:
        for i in range(args_count, width * height):
            # define subplot
            plt.subplot(width * 110 + 1 + i)
            # plot raw pixel data
            plt.imshow(random.choice(dataset[0]).reshape([28, 28]), cmap=plt.get_cmap('gray'))

    # show the figure
    plt.show()



from __future__ import annotations
from abc import ABC
import numpy as np
from EasyNN.model.abc import Model_1D
from EasyNN.typing import Array1D


class Activation(Model_1D, ABC):
    """
    Activation Models typically involve vectorized
    functions applied over the last dimension.
    """
    _parameters: Array1D = np.empty(0, dtype=float)
    _derivatives: Array1D = np.empty(0, dtype=float)

    def __setup__(self: Activation) -> None:
        pass



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    description=
        'EasyNN is a python package designed to provide an easy-to-use neural '
        'network. The package is designed to work right out of the box, while '
        'also allowing the user to customize features as they see fit.',
)
